#ifndef CORE_TEST_H
#define CORE_TEST_H

#include "core/Logger.hpp"

namespace Core {
    void finalizeTests(void);
    bool addTestResult(const char* file, bool comparison);

    template<typename T>
    bool testEqual(
        const char* file, int line, const T& wanted, const T& actual) {
        file = getShortFileName(file);
        if(addTestResult(file, wanted == actual)) {
            return true;
        }
        char buffer[512];
        formatBuffer(
            buffer, sizeof(buffer),
            TERMINAL_RED "#:# - expected '#' got '#'" TERMINAL_RESET, file,
            line, wanted, actual);
        puts(buffer);
        return false;
    }

    bool testString(
        const char* file, int line, const char* wanted, const char* actual);

    template<typename T>
    bool testString(
        const char* file, int line, const char* wanted, const T& t) {
        char buffer[512];
        toString(t, buffer, sizeof(buffer));
        return testString(file, line, wanted, static_cast<const char*>(buffer));
    }

    bool testFloat(
        const char* file, int line, float wanted, float actual, float error);
    bool testNull(const char* file, int line, const void* p);
    bool testNotNull(const char* file, int line, const void* p);

#define TEST_FLOAT(wanted, actual, error)                      \
    Core::testFloat(__FILE__, __LINE__, wanted, actual, error)

#define TEST(wanted, actual)                                  \
    Core::testEqual<Core::RemoveReference<decltype(actual)>>( \
        __FILE__, __LINE__, wanted, actual)
#define TEST_STRING(wanted, actual)                      \
    Core::testString(__FILE__, __LINE__, wanted, actual)
#define TEST_FALSE(actual) TEST(false, actual)
#define TEST_TRUE(actual) TEST(true, actual)
#define TEST_NULL(actual) Core::testNull(__FILE__, __LINE__, actual)
#define TEST_NOT_NULL(actual) Core::testNotNull(__FILE__, __LINE__, actual)
}

#endif
